import json
import requests

from requests.exceptions import RequestException

from imagefy.exceptions import ImagefyOperationException

from imagefy.models import Image, ModifiedImage

def check_tags(tags):
    """
    Tags parser for Imagefy API. Takes an iterable list as an argument and produces
    a string with the tags separated with colons.
    :param tags: The tags list.
    """
    if tags is not None:
        return ', '.join(tags)
    else:
        raise ValueError('Invalid tag list.')


class Imagefy(object):

    """
    Main Imagefy client.
    """

    def __init__(self, host, api_key=None):
        """
        Create a new connection to Imagefy.
        :param host: Imagefy server address (and port).
        :param api_key: The API key used for authentication.
        """
        self.host = host
        self.api_key = api_key

        if not host:
            raise TypeError('You have to specify the host.')

        if not api_key:
            raise TypeError('You have to specify the API key.')

    def upload(self, source_image, tags=None):
        """
        Uploads an image to the Imagefy server using the POST method.
        If successful, returns 201 CREATED with an instance saved image.
        :param source_image: The image to be sent.
        :param tags: Optional image tags.
        :return: an instance of the saved image.
        """

        url = self.host
        headers = {
            'Authorization': 'Token %s' % self.api_key
        }

        with open(source_image.name, 'rb') as image:
            files = {
                'image': image
            }
            data = {}
            if tags:
                data['tags'] = check_tags(tags)

            try:
                response = requests.post(url=url, headers=headers, data=data, files=files)
                if response.status_code == 201:
                    data = response.json()
                    return Image(data['id'], data['image'], data['created'],
                                 data['format'], data['width'], data['height'],
                                 data['user'], data['tags'])
            except RequestException as e:
                raise ImagefyOperationException("Network error.", 'upload', original_exc=e)

    def modify(self, uuid, actions):
        """
        Used to request modified images of the original image.
        :param args:
        :param kwargs:
        :return:
        """
        url = '%s%s/modified/' % (self.host, uuid)
        headers = {
            'Authorization': 'Token %s' % self.api_key,
            'Content-Type': 'application/json'
        }

        try:
            actions = json.dumps(actions)
            response = requests.post(url, headers=headers, data=actions)
            if response.status_code == 201:
                data = response.json()
                return ModifiedImage(data['id'], data['image'], data['created'],
                                     data['format'], data['width'], data['height'],
                                     data['source'], data['actions'], data['actions_hash'])
        except RequestException as e:
            raise ImagefyOperationException("Network error.", 'modification', original_exc=e)

    def retrieve(self, uuid, modified=None):
        """
        Fetches an image from the Imagefy API.
        :param uuid: An UUID of the (original) image.
        :param modified: An optional UUID for the modified version of the image.
        """
        url = '%s%s/' % (self.host, uuid)
        if modified is not None:
            url += 'modified/%s/' % modified
        headers = {
            'Authorization': 'Token %s' % self.api_key
        }

        try:
            response = requests.get(url=url, headers=headers)
            if response.status_code == 200:
                data = response.json()
                if modified is not None:
                    return ModifiedImage(data['id'], data['image'], data['created'],
                                         data['format'], data['width'], data['height'],
                                         data['source'], data['actions'], data['actions_hash'])
                return Image(data['id'], data['image'], data['created'],
                             data['format'], data['width'], data['height'],
                             data['user'], data['tags'])
        except RequestException as e:
            raise ImagefyOperationException("Network error.", 'retrieval', original_exc=e)

    def retrieve_metadata(self, uuid):
        """
        Fetches an image's metadata from the Imagefy API.
        :param uuid: An UUID of the image.
        """
        url = '%s%s/meta/' % (self.host, uuid)
        headers = {
            'Authorization': 'Token %s' % self.api_key
        }

        try:
            response = requests.get(url=url, headers=headers)
            if response.status_code == 200:
                data = response.json()
                return data
        except RequestException as e:
            raise ImagefyOperationException("Network error.", 'metadata', original_exc=e)

    def delete(self, uuid):
        """
        Deletes an uploaded image (and its modified images)
        from the Imagefy API.
        :param uuid: An UUID of the image.
        """
        url = '%s%s/' % (self.host, uuid)
        headers = {
            'Authorization': 'Token %s' % self.api_key
        }

        try:
            response = requests.delete(url=url, headers=headers)
            if response.status_code == 204:
                return True
        except RequestException as e:
            raise ImagefyOperationException("Network error.", 'deletion', original_exc=e)



from aniso8601 import parse_datetime


class Image(object):
    def __init__(self, uuid, image, created, format, width, height, user=None, tags=None):
        """
        A class representing an image object returned from the Imagefy API.
        :param uuid: The image's UUID, used to fetch modified versions of the image.
        :param url: The url from which the original file can be retrieved.
        :param created: The time and date for the image upload.
        :param tags: Optional list of tags.
        :return: An image instance.
        """
        self.uuid = uuid
        self.image = image
        self.created = parse_datetime(created)
        self.format = format
        self.width = width
        self.height = height
        self.user = user
        self.tags = tags


class ModifiedImage(object):
    def __init__(self, uuid, image, created, format, width, height, source, actions, actions_hash):
        """
        A class representing a modified image returned from the Imagefy API.
        :param uuid: The modified image's UUID.
        :param image: The url from which the modified image can be retrieved.
        :param created: The time and date for the image upload.
        :param source: The original image's url.
        :param settings: The settings used to modify the original image.
        :return: A modified image instance.
        """
        self.uuid = uuid
        self.image = image
        self.created = parse_datetime(created)
        self.format = format
        self.width = width
        self.height = height
        self.source = source
        self.actions = actions
        self.actions_hash = actions_hash



VALID_WIDTH_AND_HEIGHT = (128, 2560)
VALID_BLUR_VALUE = (0, 100)
VALID_ROTATE_VALUE = (-360, 360)
VALID_ASPECT_RATIO_VALUE = (0, 100)
VALID_CROPM_X_AND_Y = (0, 10240)
VALID_CROP_POSITIONS = ['center', 'topright', 'bottomright', 'bottomleft', 'topleft']


class BaseFilter(object):
    def get_action(self):
        raise NotImplementedError('.get_action() must be implemented in subclass')


class BlackWhite(BaseFilter):
    def __init__(self):
        self.name = 'blackwhite'

    def get_action(self):
        return {
            'name': self.name
        }


class Sepia(BaseFilter):
    def __init__(self):
        self.name = 'sepia'

    def get_action(self):
        return {
            'name': self.name
        }


class Flip(BaseFilter):
    def __init__(self):
        self.name = 'flip'

    def get_action(self):
        return {
            'name': self.name
        }


class Mirror(BaseFilter):
    def __init__(self):
        self.name = 'mirror'

    def get_action(self):
        return {
            'name': self.name
        }


class Blur(BaseFilter):
    def __init__(self, value):
        min, max = VALID_BLUR_VALUE
        if value < min or value > max:
            raise ValueError('Invalid blur value (0-100 % accepted).')
        self.name = 'blur'
        self.value = value

    def get_action(self):
        return {
            'name': self.name,
            'params': {
                'value': self.value
            }
        }


class Rotate(BaseFilter):
    def __init__(self, value):
        min, max = VALID_ROTATE_VALUE
        if value < min or value > max:
            raise ValueError('Invalid rotation angle (-360 - 360 degrees accepted).')
        self.name = 'rotate'
        self.value = value

    def get_action(self):
        return {
            'name': self.name,
            'params': {
                'value': self.value
            }
        }


class Scale(BaseFilter):
    def __init__(self, width, height):
        min, max = VALID_WIDTH_AND_HEIGHT
        if width < min or width > max or height < min or height > max:
            raise ValueError('Invalid width or height (128 - 2560 px accepted).')
        self.name = 'scale'
        self.width = width
        self.height = height

    def get_action(self):
        return {
            'name': self.name,
            'params': {
                'width': self.width,
                'height': self.height
            }
        }


class Resize(BaseFilter):
    def __init__(self, width, height):
        min, max = VALID_WIDTH_AND_HEIGHT
        if width < min or width > max or height < min or height > max:
            raise ValueError('Invalid width or height (128 - 2560 px accepted).')
        self.name = 'resize'
        self.width = width
        self.height = height

    def get_action(self):
        return {
            'name': self.name,
            'params': {
                'width': self.width,
                'height': self.height
            }
        }


class Crop(BaseFilter):
    def __init__(self, width, height, position=None):
        min, max = VALID_WIDTH_AND_HEIGHT
        if width < min or width > max or height < min or height > max:
            raise ValueError('Invalid width or height (128 - 2560 px accepted).')
        if position and position not in VALID_CROP_POSITIONS:
            raise ValueError('Invalid position ("center", "topright", "bottomright", "bottomleft", "topleft" accepted).')
        self.name = 'crop'
        self.width = width
        self.height = height
        self.position = position or 'center'

    def get_action(self):
        return {
            'name': self.name,
            'params': {
                'width': self.width,
                'height': self.height,
                'type': self.position
            }
        }


class ManualCrop(BaseFilter):
    def __init__(self, width, height, x, y):
        min, max = VALID_WIDTH_AND_HEIGHT
        min_x_y, max_x_y = VALID_CROPM_X_AND_Y
        if width < min or width > max or height < min or height > max:
            raise ValueError('Invalid width or height (128 - 2560 px accepted).')
        if x < min_x_y or x > max_x_y or y < min_x_y or y > max_x_y:
            raise ValueError('Invalid starting point (0 - 10240 px accepted).')
        self.name = 'crop'
        self.width = width
        self.height = height
        self.x = x
        self.y = y

    def get_action(self):
        return {
            'name': self.name,
            'params': {
                'width': self.width,
                'height': self.height,
                'x': self.x,
                'y': self.y
            }
        }


class AspectRatio(BaseFilter):
    def __init__(self, width, height, position=None):
        min, max = VALID_ASPECT_RATIO_VALUE
        if width < min or width > max or height < min or height > max:
            raise ValueError('Invalid aspect ratio (0 - 100 accepted).')
        if position and position not in VALID_CROP_POSITIONS:
            raise ValueError('Invalid position ("center", "topright", "bottomright", "bottomleft", "topleft" accepted).')
        self.name = 'aspect'
        self.width = width
        self.height = height
        self.position = position or 'center'

    def get_action(self):
        return {
            'name': self.name,
            'params': {
                'width': self.width,
                'height': self.height,
                'type': self.position
            }
        }



class ImagefyException(Exception):
    pass


class ImagefyOperationException(ImagefyException):
    def __init__(self, message, operation, original_exc=None):
        super(ImagefyOperationException, self).__init__(message)
        self.operation = operation
        self.original_exc = original_exc


import re

from setuptools import setup

# Fetch version number from the client init file.
version = None
for line in open('./imagefy/__init__.py'):
    m = re.search('__version__\s*=\s*(.*)', line)
    if m:
        version = m.group(1).strip()[1:-1]  # Quotes
        break
assert version

setup(
    name='imagefy',
    version=version,
    description='A Python wrapper for the Imagefy API.',
    author='Atte Valtonen',
    author_email='atte.valtonen@sin.ga',
    url='https://bitbucket.org/singster/imagefy-sdk-python',
    license='MIT',
    packages=['imagefy'],
    include_package_data=False,
    use_2to3=False,
    install_requires=[
        'aniso8601>=1.0.0',
        'requests>=2.4',
    ]
)
